unit uCom;

interface

type
  TTimer = object
    constructor         Init;
    destructor          Free;
    procedure           Delay( t:word );
    procedure           ShortDelay;
  end;

  TBit  = 0..1;
  TBit6 = $00..$3F;
  TBit14 = $0000..$3FFF;
  PCom = ^TCom;
  TCom = object
  private
    m_wCOM        : word;
    m_wBase       : word;
    m_btMCR       : byte;
    m_bPower      : boolean;
    m_Initialized : boolean;
  public
    m_Timer       : TTimer;

  private
    procedure           Initialize;
    procedure           SetRB6( bRB6:TBit );
    procedure           SetRB7( bRB7:TBit );
    procedure           SetRB6n7( bRB6,bRB7:TBit );
    function            GetRB7 : TBit;
    function            SendBit( bit:TBit ) : TBit;
  public
    constructor         Init( wComNo:word );
    destructor          Free;
    function            ComNo : word;
    procedure           SetPower( bPow:boolean );
    procedure           SendCommand( bit6:TBit6 );
    procedure           SendWord( w:TBit14 );
    function            RecvWord : TBit14;
  end;

implementation

procedure cli; assembler;
asm
{  mov   ax,$1681}
{  int   $2F}
  cli
end;

procedure sti; assembler;
asm
  sti
{  mov   ax,$1682}
{  int   $2F}
end;

constructor TTimer.Init;
begin
  Port[$43] := $96;                               { R/W LSB only          }
  Port[$61] := port[$61] or 1;                    { Start timer 2         }
  Port[$42] := 0;                                 { Initialize timer 2    }
end;

destructor  TTimer.Free;
begin
  Port[$61] := Port[$61] and not 1;
end;

procedure TTimer.Delay(t: word); assembler; { Delay / 50us          }
asm
    in      al,42h                          { Read timer 2 LSB      }
    mov     ah,al                           { Save in AH            }
    mov     cx,t                            { cx = t                }
    jmp     @2                              { I/O delay             }
@2: sub     ah,120                          { Calculate stoptime    }
@1: jmp     @0                              { I/O delay             }
@0: in      al,42h                          { Read timer 2 LSB      }
    cmp     al,ah                           { Compare with stoptime }
    jns     @1                              { Continue until timeout}
    loop    @2                              { Loop t times          }
end;

procedure TTimer.ShortDelay; assembler;
const
  t2u5 = -5;                                    { (5+1)*419ns = 2.51us  }
  t5u  = -11;                                   { 12 * 419ns  = 5.03us  }
  t7u5 = -17;                                   { 18 * 419ns  = 7.54us  }
  t10u = -23;                                   { 24 * 419ns  = 10.06us }
  sdt  = t7u5;                                  { Shortdelay  = 7.5us   }
asm
    IN      AL,42h          { Read timer 2, LSB     }
    MOV     BL,sdt          { Setup delay time      }
    ADD     BL,AL           { Calculate stop time   }
    JMP     @1              { I/O delay             }
@1: JMP     @2              { I/O delay             }
@2: IN      AL,42h          { Read timer 2, LSB     }
    CMP     AL,BL           { Compare with stoptime }
    JNS     @1              { Continue until timeout}
end;


constructor TCom.Init( wComNo:word );
const
  aCOMBases   : array[1..4] of word     = (   $3F8, $2F8, $3E8, $2E8  );
begin
  m_wCOM  := wComNo;
  m_wBase := aCOMBases[m_wCOM];

  m_btMCR := 0;
  m_bPower:= false;
  m_Initialized := false;
  m_Timer.Init;
end;

destructor TCom.Free;
begin
  if m_Initialized then
    SetPower(false);
  SetRB6n7(0,0);
  m_Timer.Free;
end;


procedure           TCom.Initialize;
begin
  m_Initialized := true;

  Port[m_wBase+4] := m_btMCR;     { Modem Control Register    }

  Port[m_wBase+3] := $03;         { Line Control Register     }
  Port[m_wBase+2] := $06;         {                           }
  Port[m_wBase+1] := $00;         { Interrupt Enable Register }

  Port[m_wBase+3] := $80;         { Line Control Register     }
  Port[m_wBase+0] := $01; {115.2} { Baud Rate Divisor LSB     }
  Port[m_wBase+1] := $00;         { Baud Rate Divisor MSB     }

  Port[m_wBase+3] := $03;         { Line Control Register     }
end;


function            TCom.ComNo : word;
begin
  ComNo := m_wCOM;
end;

procedure TCom.SetPower( bPow:boolean );
var
  i : integer;
begin
  if not m_Initialized then Initialize;

  if not bPow then
  begin
    Port[m_wBase+3] := $03;
  end else
  begin

    SetRB6n7(0,0);
    m_Timer.Delay(200);                         { 10 ms }

    Port[m_wBase+3] := $43;                     { Enable Break, no reset}
    m_Timer.Delay(4000);                        { Power on Delay, 200ms }
    Port[m_wBase+3] := $03;                     { Start reset           }
    cli;                                        { Use short reset       }
    for i:=1 to 3 do                            { pulses                }
    begin                                       { to not start auto go  }
      Port[m_wBase+0] := $00;                           { Output zero   }
      repeat until boolean(Port[m_wBase+5] and $20);    { Wait on empty }
    end;
    Port[m_wBase+3] := $43;                     { Enable break, no reset}
    sti;                                        { Interrupts ok         }
  end;

{  if bPow and (not m_bPower) then Port[m_wBase+0] := $00;}
  m_bPower := bPow;
end;

procedure TCom.SetRB6( bRB6:TBit );
begin
  m_btMCR := (m_btMCR and not 2) or ((bRB6) shl 1);
  Port[m_wBase+4] := m_btMCR;
end;

procedure TCom.SetRB7( bRB7:TBit );
begin
  m_btMCR := (m_btMCR and not 1) or (bRB7);
  Port[m_wBase+4] := m_btMCR;
end;

procedure TCom.SetRB6n7( bRB6,bRB7:TBit );
begin
  m_btMCR := (m_btMCR and not 3) or ((bRB6) shl 1) or (bRB7);
  Port[m_wBase+4] := m_btMCR;
end;

function TCom.GetRB7 : TBit;
begin
  GetRB7 := (Port[m_wBase+6] shr 4) and 1;
end;

procedure TCom.SendCommand( bit6:TBit6 );
var
  i : integer;
begin
  for i:=0 to 5 do
  begin
    SendBit( (bit6 shr i) and 1 );
  end;
end;

function TCom.SendBit( bit:TBit ) : TBit;
var
  inbit : TBit;
begin
  cli;                                          { Disable interrupts    }
  SetRB6n7(1,bit);
  m_Timer.ShortDelay;                           { Wait                  }
  inbit := GetRB7;                              { Return data           }
  SetRB6n7(0,bit and inbit);
  sti;                                          { Interrupts ok         }
  SendBit := inbit;
end;

function TCom.RecvWord : TBit14;
var
  i : integer;
  w : TBit14;
begin
  w := 0;
  SendBit(1);
  for i:=0 to 13 do
  begin
    w := w or (SendBit(1) shl i);
  end;
  SendBit(1);
  RecvWord := w;
end;

procedure TCom.SendWord( w:TBit14 );
var
  i : integer;
begin
  SendBit(0);
  for i:=0 to 13 do
  begin
    SendBit((w shr i)and 1);
  end;
  SendBit(0);
end;

begin
end.