unit uDev;

interface

uses uCom, TpString;

function unhex( c:char ) : integer;

type
  TWordArray = array[0..$7FFE] of Word;
  PWordArray = ^TWordArray;
  TByteArray = array[0..$FFFE] of Byte;
  PByteArray = ^TByteArray;
  TBoolArray = array[0..$FFFE] of Boolean;
  PBoolArray = ^TBoolArray;

  TProgress = procedure( percent:word );

  PDevice = ^TDevice;
  TDevice = object
  private
    m_Com         : PCom;
  public
    m_sName       : string;

    m_progress    : TProgress;
    m_totalwork   : Longint;
    m_currWork    : Longint;


    m_memProgram  : PWordArray;
   {m_actProgram  : PBoolArray;}
    m_bcProgram   : Word;

    m_memEEPROM   : PByteArray;
   {m_actEEPROM   : PBoolArray;}
    m_bcEEPROM    : Word;

    m_memFuses    : array[0..7] of Word;
   {m_actFuses    : array[0..7] of Boolean;}

    procedure           BeginProgress( totalWork:Longint );
    procedure           EndProgress;
    procedure           Progress( deltaCurrWork:Longint );

    function            ReadProgram   : boolean;
    function            ReadEEPROM    : boolean;
    function            ReadFuses     : boolean;
    function            WriteProgram  : boolean;
    function            WriteEEPROM   : boolean;
    function            WriteFuses    : boolean;
    function            VerifyProgram : boolean;
    function            VerifyEEPROM  : boolean;
    function            VerifyFuses   : boolean;

  public
    constructor         Init( const name:string;
                              const com:PCom;
                              progr:TProgress );
    destructor          Free;

    function            SaveToFile( const FileName:string )  : boolean;
    function            LoadFromFile( const FileName:string )  : boolean;
    function            ReadAll       : boolean;
    function            WriteAll      : boolean;
    function            VerifyAll     : boolean;
    function            EraseAll      : boolean;
    function            BlankCheck    : boolean;
  end;

implementation

function unhex( c:char ) : integer;
begin
  unhex := -1;
       if ('0'<=c) and (c<='9') then unhex := Ord(c)-$30
  else if ('A'<=c) and (c<='F') then unhex := Ord(c)-$37
  else if ('a'<=c) and (c<='f') then unhex := Ord(c)-$57;
end;

constructor         TDevice.Init( const name:string;
                                  const com:PCom;
                                  progr:TProgress  );
var
  i : integer;
begin
  m_sName     := StUpcase(name);
  m_progress  := progr;
  m_totalWork   := 0;
  m_currWork    := 0;
  m_Com := com;
  m_memProgram := nil;
 {m_actProgram := nil;}
  m_memEEPROM := nil;
 {m_actEEPROM := nil;}

 {default}
  m_bcProgram := $400;
  m_bcEEPROM  := $40;

  if m_sName='PIC16F84' then
  begin
    m_bcProgram := $400;
    m_bcEEPROM  := $40;
  end else if (m_sName='PIC16F873') or (m_sName='PIC16F874') then
  begin
    m_bcProgram := $1000;
    m_bcEEPROM  := $80;
  end else if (m_sName='PIC16F876') or (m_sName='PIC16F877') then
  begin
    m_bcProgram := $2000;
    m_bcEEPROM  := $100;
  end;

  if m_bcProgram<>0 then
  begin
    GetMem( m_memProgram, m_bcProgram*sizeof(Word) );
   {GetMem( m_actProgram, m_bcProgram*sizeof(Boolean) );}
    for i:=0 to m_bcProgram-1 do
    begin
     {m_actProgram^[i] := false;}
      m_memProgram^[i] := $3FFF;
    end;
  end;

  if m_bcEEPROM<>0 then
  begin
    GetMem( m_memEEPROM, m_bcEEPROM*sizeof(Byte) );
   {GetMem( m_actEEPROM, m_bcEEPROM*sizeof(Boolean) );}
    for i:=0 to m_bcEEPROM-1 do
    begin
     {m_actEEPROM^[i] := false;}
      m_memEEPROM^[i] := $FF;
    end;
  end;

  for i:=Low(m_memFuses) to High(m_memFuses) do
  begin
   {m_actFuses[i] := false;}
    m_memFuses[i] := $3FFF;
  end;
end;

destructor          TDevice.Free;
begin
  if m_bcProgram<>0 then
  begin
    FreeMem( m_memProgram, m_bcProgram*sizeof(Word) );
   {FreeMem( m_actProgram, m_bcProgram*sizeof(Boolean) );}
  end;
  if m_bcEEPROM<>0 then
  begin
    FreeMem( m_memEEPROM, m_bcEEPROM*sizeof(Byte) );
   {FreeMem( m_actEEPROM, m_bcEEPROM*sizeof(Boolean) );}
  end;
end;

procedure           TDevice.BeginProgress( totalWork:Longint );
begin
  if m_totalWork=0 then
  begin
    m_totalWork := totalWork;
    if @m_progress<>nil then m_progress( 1001 );
    m_currWork := 0;
  end;
end;

procedure           TDevice.EndProgress;
begin
  m_totalWork := 0;
  m_currWork := 0;
  if @m_progress<>nil then m_progress( 1002 );
end;

procedure           TDevice.Progress( deltaCurrWork:Longint );
var
  val1, val2 : Longint;
begin
  if m_totalWork<>0 then
  begin
    val1 := (m_currWork*100) div m_totalWork;
    Inc( m_CurrWork, deltaCurrWork );
    val2 := (m_currWork*100) div m_totalWork;
    if val1<>val2 then
    begin
      if @m_progress<>nil then m_progress( val1 );
    end;
  end;
end;

function            TDevice.ReadProgram : boolean;
var
  i : integer;
begin
  with m_Com^ do
  begin
    SetPower(true);

    for i:=0 to m_bcProgram-1 do
    begin
      Progress(1);
      SendCommand( $04 );
      m_memProgram^[i] := RecvWord;
      SendCommand( $06 );
    end;
    SetPower(false);
  end;

  ReadProgram := true;
end;

function            TDevice.ReadFuses   : boolean;
var
  i : integer;
begin
  with m_Com^ do
  begin
    SetPower(true);
    SendCommand( $00 );
    SendWord( $3FFF );
    for i:=0 to 7 do
    begin
      Progress(1);
      SendCommand( $04 );
      m_memFuses[i] := RecvWord;
     {m_actFuses[i] := true;}
      SendCommand( $06 );
    end;
    SetPower(false);
  end;
  ReadFuses := true;
end;

function            TDevice.ReadEEPROM : boolean;
var
  i : integer;
begin
  with m_Com^ do
  begin
    SetPower(true);
    for i:=0 to m_bcEEPROM-1 do
    begin
      Progress(1);
      SendCommand( $05 );
      m_memEEPROM^[i] := RecvWord;
      SendCommand( $06 );
    end;
    SetPower(false);
  end;
  ReadEEPROM := true;
end;

function            TDevice.WriteProgram  : boolean;
var
  i : integer;
begin
  WriteProgram := true;
  with m_Com^ do
  begin
    SetPower(true);
    for i:=0 to m_bcProgram-1 do
    begin
      Progress(1);
      SendCommand( $04 );
      if RecvWord <> m_memProgram^[i] then
      begin
        SendCommand( $02 );
        SendWord( m_memProgram^[i] );
        SendCommand( $08 );
        m_Timer.Delay(200);                         { 10 ms }
        SendCommand( $04 );
        if RecvWord <> m_memProgram^[i] then
        begin
          WriteProgram := false;
          {Exit;}
        end;
      end;
      SendCommand( $06 );
    end;
    SetPower(false);
  end;
end;

function            TDevice.WriteFuses    : boolean;
var
  i : integer;
begin
  WriteFuses := true;
  with m_Com^ do
  begin
    SetPower(true);
    SendCommand( $00 );
    SendWord( $3FFF );
    for i:=0 to 7 do
    begin
      Progress(1);
      SendCommand( $04 );
      if RecvWord <> m_memFuses[i] then
      begin
        SendCommand( $02 );
        SendWord( m_memFuses[i] );
        SendCommand( $08 );
        m_Timer.Delay(200);                         { 10 ms }
        SendCommand( $04 );
        if RecvWord <> m_memFuses[i] then
        begin
          if (i<>4) and (i<>5) and (i<>6) then { skip device id }
            WriteFuses := false;
          {Exit;}
        end;
      end;
      SendCommand( $06 );
    end;
    SetPower(false);
  end;
end;

function            TDevice.WriteEEPROM   : boolean;
var
  i : integer;
begin
  WriteEEPROM := true;
  with m_Com^ do
  begin
    SetPower(true);
    for i:=0 to m_bcEEPROM-1 do
    begin
      Progress(1);
      SendCommand( $05 );
      if Lo(RecvWord) <> Lo(m_memEEPROM^[i]) then
      begin
        SendCommand( $03 );
        SendWord( m_memEEPROM^[i] );
        SendCommand( $08 );
        m_Timer.Delay(200);                         { 10 ms }
        SendCommand( $05 );
        if Lo(RecvWord) <> Lo(m_memEEPROM^[i]) then
        begin
          WriteEEPROM := false;
          {Exit;}
        end;
      end;
      SendCommand( $06 );
    end;
    SetPower(false);
  end;
end;


function            TDevice.VerifyProgram : boolean;
var
  i : integer;
label
  __not_equal;
begin
  VerifyProgram := false;

  with m_Com^ do
  begin
    SetPower(true);

    for i:=0 to m_bcProgram-1 do
    begin
      Progress(1);
      SendCommand( $04 );
      if m_memProgram^[i]<>RecvWord then goto __not_equal;
      SendCommand( $06 );
    end;
    SetPower(false);
  end;
  VerifyProgram := true;

__not_equal:
  m_Com^.SetPower(false);
end;


function            TDevice.VerifyEEPROM  : boolean;
var
  i : integer;
label
  __not_equal;
begin
  VerifyEEPROM := false;

  with m_Com^ do
  begin
    SetPower(true);
    for i:=0 to m_bcEEPROM-1 do
    begin
      Progress(1);
      SendCommand( $05 );
      if Lo(m_memEEPROM^[i])<>Lo(RecvWord) then goto __not_equal;
      SendCommand( $06 );
    end;
    SetPower(false);
  end;
  VerifyEEPROM := true;
__not_equal:
  m_Com^.SetPower(false);
end;

function            TDevice.VerifyFuses   : boolean;
var
  i : integer;
label
  __not_equal;
begin
  VerifyFuses := false;

  with m_Com^ do
  begin
    SetPower(true);
    SendCommand( $00 );
    SendWord( $3FFF );
    for i:=0 to 7 do
    begin
      Progress(1);
      SendCommand( $04 );
      if m_memFuses[i]<>RecvWord then
        if (i<>4) and (i<>5) and (i<>6) then { skip device id }
          goto __not_equal;
      SendCommand( $06 );
    end;
    SetPower(false);
  end;

  VerifyFuses  := true;
__not_equal:
  m_Com^.SetPower(false);
end;


function            TDevice.EraseAll : boolean;
var
  i : integer;
begin
  EraseAll := true;
  BeginProgress( 3  );
  with m_Com^ do
  begin
    SetPower(true);
    SendCommand( $00 ); SendWord( $3FFF );
    for i:=1 to 7 do SendCommand( $06 );
    SendCommand( $02 ); SendWord( $3FFF );
    SendCommand( $01 );
    SendCommand( $07 );
    SendCommand( $08 );
    Progress(1);
    m_Timer.Delay(200);                         { 10 ms }
    Progress(1);
    SendCommand( $01 );
    SendCommand( $07 );
    SetPower(false);
  end;
  EndProgress;
end;

function            TDevice.BlankCheck : boolean;
var
  i  : Word;
label
  __not_blank;
begin
  BlankCheck := false;
  BeginProgress( 8+m_bcProgram+m_bcEEPROM );

  with m_Com^ do
  begin

    SetPower(true);
    SendCommand( $00 );
    SendWord( $3FFF );
    for i:=0 to 7 do
    begin
      Progress(1);
      SendCommand( $04 );
      if RecvWord<>$3FFF then
        if (i<>4) and (i<>5) and (i<>6) then { skip device id }
          goto __not_blank;
      SendCommand( $06 );
    end;
    SetPower(false);

    SetPower(true);
    for i:=0 to m_bcProgram-1 do
    begin
      Progress(1);
      SendCommand( $04 );
      if RecvWord<>$3FFF then goto __not_blank;
      SendCommand( $06 );
    end;
    SetPower(false);

    SetPower(true);
    for i:=0 to m_bcEEPROM-1 do
    begin
      Progress(1);
      SendCommand( $05 );
      if Lo(RecvWord)<>$FF then goto __not_blank;
      SendCommand( $06 );
    end;
    SetPower(false);

  end;

  BlankCheck := true;

__not_blank:
  m_Com^.SetPower(false);
  EndProgress;
end;


function            TDevice.ReadAll     : boolean;
var
  bSuccess : boolean;
begin
  BeginProgress( 8+m_bcProgram+m_bcEEPROM );

  bSuccess := ReadFuses;
  bSuccess := ReadProgram and bSuccess;
  bSuccess := ReadEEPROM and bSuccess;
  ReadAll := bSuccess;

  EndProgress;
end;


function            TDevice.WriteAll : boolean;
var
  bSuccess : boolean;
begin
  BeginProgress( 8+m_bcProgram+m_bcEEPROM );

  bSuccess := WriteProgram;
  bSuccess := WriteEEPROM and bSuccess;
  { ᫥, ..   CP,  ⮣  㤠 祣  }
  bSuccess := WriteFuses and bSuccess;
  WriteAll := bSuccess;

  EndProgress;
end;

function            TDevice.VerifyAll : boolean;
var
  bSuccess : boolean;
begin
  BeginProgress( 8+m_bcProgram+m_bcEEPROM );

  bSuccess := VerifyFuses;
  if bSuccess then bSuccess := VerifyProgram and bSuccess;
  if bSuccess then bSuccess := VerifyEEPROM and bSuccess;
  VerifyAll := bSuccess;

  EndProgress;
end;

function            TDevice.SaveToFile( const FileName:string )  : boolean;
  function HexLineW( p:PWordArray; off,bc:word ) : string;
  var
    s : string;
    bChk : byte;
    i : word;
  begin
    s := ':' + HexB(Lo(bc)*2) + HexW(off) + '00';
    bChk := 0;
    Dec(bChk,Lo(bc)*2);
    Dec(bChk,Hi(off));
    Dec(bChk,Lo(off));
    for i:=0 to Lo(bc)-1 do
    begin
      Dec(bChk,Lo(p^[i]));
      Dec(bChk,Hi(p^[i]));
      s := s + HexB(Lo(p^[i])) + HexB(Hi(p^[i]));
    end;
    HexLineW := s + HexB(bChk);
  end;
  function HexLineB( p:PByteArray; off,bc:word ) : string;
  var
    s : string;
    bChk : byte;
    i : word;
  begin
    s := ':' + HexB(Lo(bc)*2) + HexW(off) + '00';
    bChk := 0;
    Dec(bChk,Lo(bc)*2);
    Dec(bChk,Hi(off));
    Dec(bChk,Lo(off));
    for i:=0 to Lo(bc)-1 do
    begin
      Dec(bChk,p^[i]);
      s := s + HexB(p^[i]) + '00';
    end;
    HexLineB := s + HexB(bChk);
  end;
var
  i : word;
  f : Text;
begin
  SaveToFile := false;

  Assign(F,FileName);
{$I-}
  Rewrite(F);
  if IOResult<>0 then Exit;
{$I+}

  for i:=0 to (m_bcProgram div $08)-1 do
    WriteLn( F, HexLineW( Pointer(@m_memProgram^[$08*i]), $0000+$10*i, $08 ) );

  for i:=0 to (m_bcEEPROM div $08)-1 do
    WriteLn( F, HexLineB( Pointer(@m_memEEPROM^[$08*i]), $4200+$10*i, $08 ) );

  WriteLn( F, HexLineW( Pointer(@m_memFuses[$00]), $4000, $04 ) );
  WriteLn( F, HexLineW( Pointer(@m_memFuses[$07]), $400E, $01 ) );

  WriteLn( F, ':00000001FF');
  Close(F);

  SaveToFile := true;
end;

function            TDevice.LoadFromFile( const FileName:string )  : boolean;

  function Hex2B( const s:string; var b:Byte ) : boolean;
  var
    i : integer;
  begin
    Hex2B := false;
    if Length(s)<>2 then Exit;
    i := unhex(s[1]); if i<0 then Exit; b := i shl 4;
    i := unhex(s[2]); if i<0 then Exit; b := b or i;
    Hex2B := true;
  end;

  function Hex2W( const s:string; var w:Word ) : boolean;
  var
    i : integer;
  begin
    Hex2W := false;
    if Length(s)<>4 then Exit;
    i := unhex(s[1]); if i<0 then Exit; w := i shl 12;
    i := unhex(s[2]); if i<0 then Exit; w := w or (i shl 8);
    i := unhex(s[3]); if i<0 then Exit; w := w or (i shl 4);
    i := unhex(s[4]); if i<0 then Exit; w := w or i;
    Hex2W := true;
  end;
var
  i : word;
  f : Text;
  S : string;
  hxLen : byte;
  hxOff : word;
  hxTag : byte;
  hxData : array[0..$40] of byte;
  hxChk : byte;
label
  __failed;
begin
  LoadFromFile := false;

  Assign(F,FileName);
{$I-}
  Reset(F);
  if IOResult<>0 then Exit;
{$I+}

  repeat
    ReadLn(F,S);
    if Length(S)=0 then break;
    if (Length(S)<11) or (S[1]<>':') then goto __failed;

    if not Hex2B( S[2]+S[3], hxLen ) then goto __failed;
    if Length(S)<1+(1+2+1+hxLen+1)*2 then goto __failed;
    if not Hex2W( S[4]+S[5]+S[6]+S[7], hxOff ) then goto __failed;
    if not Hex2B( S[8]+S[9], hxTag ) then goto __failed;
    if hxLen<>0 then
      for i:=0 to hxLen-1 do
        if not Hex2B( S[10+i*2]+S[11+i*2], hxData[i] ) then goto __failed;
    if not Hex2B( S[10+2*hxLen]+S[11+2*hxLen], hxChk ) then goto __failed;

    Inc(hxChk,hxLen);
    Inc(hxChk,Lo(hxOff));
    Inc(hxChk,Hi(hxOff));
    Inc(hxChk,hxTag);
    if hxLen<>0 then
      for i:=0 to hxLen-1 do
        Inc(hxChk,hxData[i]);
    if hxChk<>0 then goto __failed;

    if hxLen>1 then
    begin
      if hxOff<m_bcProgram*2 then
      begin
        for i:=0 to (hxLen div 2)-1 do
          if ((hxOff-$0000)div 2)+i < m_bcProgram then
            m_memProgram^[((hxOff-$0000)div 2)+i] := hxData[2*i] or (hxData[2*i+1] shl 8);
      end else if (hxOff-$4000)<8*2 then
      begin
        for i:=0 to (hxLen div 2)-1 do
          if ((hxOff-$4000)div 2)+i < 8 then
            m_memFuses[((hxOff-$4000)div 2)+i] := hxData[2*i] or (hxData[2*i+1] shl 8);
      end else if (hxOff-$4200)<m_bcEEPROM*2 then
      begin
        for i:=0 to (hxLen div 2)-1 do
          if ((hxOff-$4200)div 2)+i < m_bcEEPROM then
            m_memEEPROM^[((hxOff-$4200)div 2)+i] := hxData[2*i];
      end;
    end;

  until false;

  LoadFromFile := true;
__failed:
  Close(F);
end;


end.